"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const request = require("request-promise");
const entityDecode = require("decode-html");
const crplData = require("./crpl-data.json");
class ParseTree extends Array {
    push(...branches) {
        branches.forEach(branch => branch.parent = this);
        return super.push(...branches);
    }
    splice(start, deleteCount, ...items) {
        items.forEach(branch => branch.parent = this);
        return super.splice(start, deleteCount, ...items);
    }
}
exports.ParseTree = ParseTree;
var ParseMode;
(function (ParseMode) {
    ParseMode[ParseMode["normal"] = 0] = "normal";
    ParseMode[ParseMode["string"] = 1] = "string";
    ParseMode[ParseMode["comment"] = 2] = "comment";
})(ParseMode = exports.ParseMode || (exports.ParseMode = {}));
var CRPLType;
(function (CRPLType) {
    CRPLType[CRPLType["b"] = 1] = "b";
    CRPLType[CRPLType["i"] = 7] = "i";
    CRPLType[CRPLType["f"] = 5] = "f";
    CRPLType[CRPLType["s"] = 1] = "s";
    CRPLType[CRPLType["l"] = 9] = "l";
    CRPLType[CRPLType["n"] = 1] = "n";
    CRPLType[CRPLType["o"] = 0] = "o";
    CRPLType[CRPLType["listBit"] = 16] = "listBit";
    CRPLType[CRPLType["bb"] = 17] = "bb";
    CRPLType[CRPLType["ii"] = 23] = "ii";
    CRPLType[CRPLType["ff"] = 21] = "ff";
    CRPLType[CRPLType["ss"] = 17] = "ss";
    CRPLType[CRPLType["ll"] = 25] = "ll";
    CRPLType[CRPLType["nn"] = 17] = "nn";
    CRPLType[CRPLType["shrug"] = 255] = "shrug"; // for when we don't know what's happening and have decided we don't care
})(CRPLType = exports.CRPLType || (exports.CRPLType = {}));
exports.sigSpec = {
    convertible(t1, t2) { return (t1 & t2) === t2; },
    isList(t) { return t && CRPLType.listBit; },
    inTypes: new Map([
        ['b', CRPLType.b], ['i', CRPLType.i], ['f', CRPLType.f],
        ['x', CRPLType.f], ['y', CRPLType.f], ['z', CRPLType.f],
        ['fi', CRPLType.f],
        ['s', CRPLType.s],
        ['l', CRPLType.l],
        ['n', CRPLType.n],
        ['b*', CRPLType.bb], ['i*', CRPLType.ii], ['f*', CRPLType.ff],
        ['x*', CRPLType.ff], ['y*', CRPLType.ff], ['z*', CRPLType.ff],
        ['fi*', CRPLType.ff],
        ['s*', CRPLType.ss],
        ['l*', CRPLType.ll],
        ['n*', CRPLType.nn], ['*', CRPLType.nn]
    ]),
    outTypes: new Map([
        ['b', CRPLType.i], ['i', CRPLType.i], ['f', CRPLType.f],
        ['x', CRPLType.i], ['y', CRPLType.i], ['z', CRPLType.f],
        ['fi', CRPLType.f],
        ['s', CRPLType.s],
        ['l', CRPLType.l],
        ['n', CRPLType.n],
        ['b*', CRPLType.ii], ['i*', CRPLType.ii], ['f*', CRPLType.ff],
        ['x*', CRPLType.ii], ['y*', CRPLType.ii], ['z*', CRPLType.ff],
        ['fi*', CRPLType.ff],
        ['s*', CRPLType.ss],
        ['l*', CRPLType.ll],
        ['n*', CRPLType.nn], ['*', CRPLType.nn]
    ])
};
class StackDelta {
    constructor([input, output, odd], { dup = false, dup2 = false, swap = false, arithmetic = false, reflow = false, listout = false }) {
        let ins = input.length;
        let outs = output.length;
        this.raw = new ArrayBuffer(ins + outs);
        this.input = new Uint8Array(this.raw, 0, ins);
        this.output = new Uint8Array(this.raw, ins, outs);
        input.forEach((t, i) => this.input[i] = t);
        output.forEach((t, i) => this.output[i] = t);
        this.dup = dup;
        this.dup2 = dup2;
        this.swap = swap;
        this.arithmetic = arithmetic;
        this.reflow = reflow;
        this.listout = listout;
    }
    patch(stack) {
        let arithmeticType = CRPLType.i;
        this.input.reduceRight((_, expected, i) => {
            let popped = stack.pop(expected);
            if (popped === undefined) {
                throw { popped, place: i, expected };
            }
            arithmeticType &= popped;
            if (!exports.sigSpec.convertible(popped, expected)) {
                throw { popped, place: i, expected };
            }
            return _;
        }, null);
        if (this.arithmetic) {
            stack.push(...(this.output.map(t => arithmeticType | t)));
        }
        else {
            stack.push(...this.output);
        }
    }
}
exports.StackDelta = StackDelta;
class Stack extends Array {
    constructor(initial = []) {
        super(1024);
        initial.forEach((t, i) => this[i] = t);
    }
    pop(expected) {
        if (this.length) {
            let result = super.pop();
            if (result && exports.sigSpec.isList(result)) {
                this.push(result);
            }
            return result;
        }
        else {
            return 0;
        }
    }
    clear() {
        this.splice(0, this.length);
    }
    patch(delta) {
        delta.patch(this);
    }
}
exports.Stack = Stack;
class StackTracker {
    constructor(base) {
        this.base = new Stack(base);
    }
    patch(deltas) {
        let stack = new Stack(this.base);
        deltas.forEach(delta => stack.patch(delta));
    }
}
exports.StackTracker = StackTracker;
let localSignatures = new Map();
Object.entries(crplData.signatures).forEach(([token, sigWithBadType]) => {
    let sigArray = sigWithBadType;
    let reflow = /^(break|do|else|endif|endonce|endwhile|loop|once|endonce|repeat|return|while)$/.test(token);
    let arithmetic = sigArray[0][0] === 'fi';
    let dup = token === 'dup';
    let dup2 = token === 'dup2';
    let swap = token === 'swap';
    let listout = /^(get.*unitsinrange|getcoreswithvar)/.test(token);
    let deltaArray = [
        sigArray[0].map(t => exports.sigSpec.inTypes.get(t)),
        sigArray[1].map(t => exports.sigSpec.outTypes.get(t)),
        sigArray[2]
    ];
    if (/^(clearstack|(pre|ap)pendstacktolist)$/.test(token)
        || token[0] === ':') {
        return localSignatures.set(token, new StackTracker(deltaArray[1]));
    }
    localSignatures.set(token, new StackDelta(deltaArray, { dup, dup2, swap, arithmetic, reflow, listout }));
});
exports.signatures = localSignatures;
function scrapeSignatures(groupSize = 10) {
    return __awaiter(this, void 0, void 0, function* () {
        let groups = [];
        let group;
        crplData.words.forEach((word, i) => {
            if (!(i % groupSize)) {
                group = [];
                groups.push(group);
            }
            group.push(word);
        });
        let result = [];
        for (let i = 0; i < groups.length; i++) {
            console.log(i + 1, '/', groups.length);
            result.push(yield Promise.all(groups[i].map((word) => __awaiter(this, void 0, void 0, function* () {
                let item = {};
                try {
                    let docs = yield request(crplData.prefix + word + crplData.suffix);
                    docs = docs.replace(/\r/g, '');
                    docs = /=====(.*\n)*/.exec(docs)[0];
                    docs = docs.replace(/(===+\s*Examples|<\/textarea)(.*\n)*/, '');
                    docs = docs.replace(/''|%%/g, '`');
                    docs = entityDecode(docs);
                    let prepresig = /\| *`? *((?:[bifxysln]\d? *?)*) *-+ *((?:[bifxysln]\d? *?)*) *`? *\|$/i.exec(docs);
                    if (!prepresig) {
                        throw new Error(`Error getting signature from ${word}`);
                    }
                    let presig = prepresig.slice(1, 3);
                    let signature = [
                        presig[0] ? presig[0].split(/ +/).map(t => t[0].toLowerCase()) : [],
                        presig[1] ? presig[1].split(/ +/).map(t => t[0].toLowerCase()) : []
                    ];
                    item[word] = signature;
                    return item;
                }
                catch (e) {
                    console.log(word, e);
                    item[word] = [['ERROR'], []];
                    return item;
                }
            }))));
        }
        console.log(JSON.stringify(result, null, ' ')
            .replace(/\n  ( +|(?=}))/g, '')
            .replace(/\n [\[\]],?/g, '')
            .replace(/},?\n  {/g, ',\n'));
    });
}
exports.scrapeSignatures = scrapeSignatures;
exports.stackSpec = {
    up: ['(', 'define', ':', 'func', 'do', 'once', 'if', 'else', 'while', 'repeat', 'comment', '"'],
    down: new Map([
        [')', '('],
        ['define', 'start'], [':', 'define'], ['value', ':'],
        ['func', /^(func|start|bottom)$/],
        ['loop', 'do'],
        ['endonce', 'once'],
        ['else', 'if'],
        ['endif', /^(if|else)$/],
        ['repeat', 'while'],
        ['endwhile', 'repeat'],
        ['endcomment', 'comment'],
        ['"', '"']
    ]),
    unmatched: new Map([
        // number represents where the start of the block is
        // inside the ParseTree in relation to the error'd token
        // 0 = same token, eg ["once", [...]] -> "once" errors, its block starts at itself
        // 2 = different token, eg ["if", [...], "else", [...]] -> "else" errors, its block starts at "if"
        ['(', 0],
        ['do', 0],
        ['once', 0],
        ['if', 0],
        ['else', 2],
        ['while', 0],
        ['repeat', 2],
        ['comment', 0],
        ['"', 0]
    ]),
};
function surely(x) { return x; }
exports.surely = surely;
//# sourceMappingURL=helpers.js.map